import binascii

from Crypto.Cipher import AES
from Crypto.Hash import keccak
from Crypto.PublicKey import ECC
from Crypto.Random import get_random_bytes
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from eth_keys import keys

block_size = AES.block_size
address_size = 20
func_sig_size = 4
ct_size = 32
key_size = 32


def encrypt(key, plaintext):
    # Ensure plaintext is smaller than 128 bits (16 bytes)
    if len(plaintext) > block_size:
        raise ValueError("Plaintext size must be 128 bits or smaller.")

    # Ensure key size is 128 bits (16 bytes)
    if len(key) != block_size:
        raise ValueError("Key size must be 128 bits.")

    # Create a new AES cipher block using the provided key
    cipher = AES.new(key, AES.MODE_ECB)

    # Generate a random value 'r' of the same length as the block size
    r = get_random_bytes(block_size)

    # Encrypt the random value 'r' using AES in ECB mode
    encrypted_r = cipher.encrypt(r)

    # Pad the plaintext with zeros if it's smaller than the block size
    plaintext_padded = bytes(block_size - len(plaintext)) + plaintext

    # XOR the encrypted random value 'r' with the plaintext to obtain the ciphertext
    ciphertext = bytes(x ^ y for x, y in zip(encrypted_r, plaintext_padded))

    return ciphertext, r


def decrypt(key, r, ciphertext):
    if len(ciphertext) != block_size:
        raise ValueError("Ciphertext size must be 128 bits.")

    # Ensure key size is 128 bits (16 bytes)
    if len(key) != block_size:
        raise ValueError("Key size must be 128 bits.")

    # Ensure random size is 128 bits (16 bytes)
    if len(r) != block_size:
        raise ValueError("Random size must be 128 bits.")

    # Create a new AES cipher block using the provided key
    cipher = AES.new(key, AES.MODE_ECB)

    # Encrypt the random value 'r' using AES in ECB mode
    encrypted_r = cipher.encrypt(r)

    # XOR the encrypted random value 'r' with the ciphertext to obtain the plaintext
    plaintext = bytes(x ^ y for x, y in zip(encrypted_r, ciphertext))

    return plaintext


def load_aes_key(file_path):
    # Read the hex-encoded contents of the file
    with open(file_path, 'r') as file:
        hex_key = file.read().strip()

    # Decode the hex string to binary
    key = binascii.unhexlify(hex_key)

    # Ensure the key is the correct length
    if len(key) != block_size:
        raise ValueError(f"Invalid key length: {len(key)} bytes, must be {block_size} bytes")

    return key


def write_aes_key(file_path, key):
    # Ensure the key is the correct length
    if len(key) != block_size:
        raise ValueError(f"Invalid key length: {len(key)} bytes, must be {block_size} bytes")

    # Encode the key to hex string
    hex_key = binascii.hexlify(key).decode()

    # Write the hex-encoded key to the file
    with open(file_path, 'w') as file:
        file.write(hex_key)


def generate_aes_key():
    # Generate a random 128-bit AES key
    key = get_random_bytes(block_size)

    return key


def generate_ECDSA_private_key():
    # Generate a new ECDSA private key
    private_key = ECC.generate(curve='P-256')

    # Get the raw bytes of the private key
    return private_key.d.to_bytes(private_key.d.size_in_bytes(), byteorder='big')


def sign_input_text(sender, addr, func_sig, ct, key):
    # Ensure all input sizes are the correct length
    if len(sender) != address_size:
        raise ValueError(f"Invalid sender address length: {len(sender)} bytes, must be {address_size} bytes")
    if len(addr) != address_size:
        raise ValueError(f"Invalid contract address length: {len(addr)} bytes, must be {address_size} bytes")
    if len(func_sig) != func_sig_size:
        raise ValueError(f"Invalid signature size: {len(func_sig)} bytes, must be {func_sig_size} bytes")
    if len(ct) != ct_size:
        raise ValueError(f"Invalid ct length: {len(ct)} bytes, must be {ct_size} bytes")
    # Ensure the key is the correct length
    if len(key) != key_size:
        raise ValueError(f"Invalid key length: {len(key)} bytes, must be {key_size} bytes")

    # Create the message to be signed by appending all inputs
    message = sender + addr + func_sig + ct

    return sign(message, key)


def sign(message, key):
    # Sign the message
    pk = keys.PrivateKey(key)
    signature = pk.sign_msg(message).to_bytes()

    return signature


def build_input_text(plaintext, user_aes_key, sender, contract, func_sig, signing_key):
    sender_address_bytes = bytes.fromhex(sender.address[2:])
    contract_address_bytes = bytes.fromhex(contract.address[2:])

    # Convert the integer to a byte slice with size aligned to 8.
    plaintext_bytes = plaintext.to_bytes((plaintext.bit_length() + 7) // 8, 'big')

    # Encrypt the plaintext with the user's AES key
    ciphertext, r = encrypt(user_aes_key, plaintext_bytes)
    ct = ciphertext + r

    # Create the function signature
    func_hash = get_func_sig(func_sig)
    # Sign the message
    signature = sign_input_text(sender_address_bytes, contract_address_bytes, func_hash, ct, signing_key)

    # Convert the ct to an integer
    int_cipher_text = int.from_bytes(ct, byteorder='big')

    return int_cipher_text, signature


def generate_rsa_keypair():
    # Generate RSA key pair
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048
    )

    # Serialize private key
    private_key_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    # Get public key
    public_key = private_key.public_key()
    # Serialize public key
    public_key_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    return private_key_bytes, public_key_bytes


def encrypt_rsa(public_key_bytes, plaintext):
    # Load public key
    public_key = serialization.load_der_public_key(public_key_bytes)
    # Encrypt plaintext
    ciphertext = public_key.encrypt(
        plaintext,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return ciphertext


def decrypt_rsa(private_key_bytes, ciphertext):
    # Load private key
    private_key = serialization.load_der_private_key(private_key_bytes, password=None)
    # Decrypt ciphertext
    plaintext = private_key.decrypt(
        ciphertext,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return plaintext


# Function to compute Keccak-256 hash
def keccak256(data):
    # Create Keccak-256 hash object
    hash_obj = keccak.new(digest_bits=256)

    # Update hash object with data
    hash_obj.update(data)

    # Compute hash and return
    return hash_obj.digest()


def get_func_sig(function_signature):
    # Convert function signature to bytes
    function_signature_bytes = function_signature.encode('utf-8')

    # Compute Keccak-256 hash on the function signature
    function_signature_bytes_hash = keccak256(function_signature_bytes)

    # Take first 4 bytes of the hash
    return function_signature_bytes_hash[:4]



import os

from eth_account import Account
from web3 import Web3
from web3.middleware import geth_poa_middleware

from coti.crypto_utils import block_size, decrypt

SOLC_VERSION = '0.8.19'


def web3_connected(web3):
    return web3.is_connected()


def print_network_details(web3):
    print('provider: ', web3.eth.w3.provider.endpoint_uri)
    print('chain-id: ', web3.eth.chain_id)
    print('latest block: ', web3.eth.get_block('latest').get('hash').hex())


def print_account_details(web3):
    print('account address:', web3.eth.default_account.address)
    print('account balance: ', get_native_balance(web3), 'wei (', web3.from_wei(get_native_balance(web3), 'ether'),
          ' ether)')
    print('account nonce: ', get_nonce(web3))


def init_web3(node_https_address, eoa, error_not_connected=True):
    web3 = Web3(Web3.HTTPProvider(node_https_address))
    web3.eth.default_account = eoa
    web3.middleware_onion.inject(geth_poa_middleware, layer=0)
    if error_not_connected:
        if not web3_connected(web3):
            raise Exception("Connection to node failed!")
    print_network_details(web3)
    print_account_details(web3)
    return web3


def get_eoa(account_private_key):
    eoa = Account.from_key(account_private_key)
    if address_valid(eoa.address):
        return eoa
    raise Exception("eoa from private key is not valid!")


def validate_address(address):
    return {'valid': Web3.is_address(address), 'safe': Web3.to_checksum_address(address)}


def get_latest_block(web3):
    return web3.eth.get_block('latest')


def get_nonce(web3):
    return web3.eth.get_transaction_count(web3.eth.default_account.address)


def get_address_valid_and_checksum(address):
    return {'valid': Web3.is_address(address), 'safe': Web3.to_checksum_address(address)}


def address_valid(address):
    return get_address_valid_and_checksum(address)['valid']


def get_native_balance(web3, address=None):
    if address is None:
        return web3.eth.get_balance(web3.eth.default_account.address)
    if address is not None and address_valid(address):
        return web3.eth.get_balance(address)
    if not address_valid(address):
        raise Exception('address ', address, ' is not valid!')


def load_contract(file_path):
    # Ensure the file path is valid
    if not os.path.exists(file_path):
        raise Exception(f"The file {file_path} does not exists")

    # Read the Solidity source code from the file
    with open(file_path, 'r') as file:
        return file.read()


def transfer_native(web3, recipient_address, private_key, amount_to_transfer_ether, native_gas_units):
    tx = {
        'to': recipient_address,
        'from': web3.eth.default_account.address,
        'value': web3.to_wei(amount_to_transfer_ether, 'ether'),  # Transaction value (0.1 Ether in this example)
        'nonce': get_nonce(web3),
        'gas': native_gas_units,  # Gas limit for the transaction
        'gasPrice': web3.eth.gas_price,
        'chainId': web3.eth.chain_id
    }
    validate_gas_estimation(web3, tx)
    tx_receipt = sign_and_send_tx(web3, private_key, tx)
    return tx_receipt


def validate_gas_estimation(web3, tx):
    valid, gas_estimate = is_gas_units_estimation_valid(web3, tx)
    if valid is False:
        raise Exception('not enough gas for tx (provided: ' + str(tx.get('gas')) + ' needed by estimation: ' + str(
            gas_estimate) + ')')


def is_gas_units_estimation_valid(web3, tx):
    estimate_gas = web3.eth.estimate_gas(tx)
    if tx['gas'] >= estimate_gas:
        return True, estimate_gas
    return False, estimate_gas


def get_function_signature(function_abi):
    # Extract the input types from the ABI
    input_types = ','.join([param['type'] for param in function_abi.get('inputs', [])])

    # Generate the function signature
    return f"{function_abi['name']}({input_types})"


def deploy_contract(contract, kwargs, tx_params):
    func = contract.constructor(**kwargs)
    return exec_func_via_transaction(func, tx_params)


def exec_func_via_transaction(func, tx_params):
    web3 = tx_params['web3']
    gas_limit = tx_params['gas_limit']
    gas_price_gwei = tx_params['gas_price_gwei']
    account_private_key = tx_params['eoa_private_key']
    tx = func.build_transaction({
        'from': web3.eth.default_account.address,
        'chainId': web3.eth.chain_id,
        'nonce': get_nonce(web3),
        'gas': gas_limit,
        'gasPrice': web3.to_wei(gas_price_gwei, 'gwei')
    })
    # validate_gas_estimation(web3, tx)
    tx_receipt = sign_and_send_tx(web3, account_private_key, tx)
    return tx_receipt


def sign_and_send_tx(web3, private_key, transaction):
    try:
        signed_tx = web3.eth.account.sign_transaction(transaction, private_key)
    except Exception as e:
        raise Exception(f"Failed to sign the transaction: {e}")
    try:
        tx_hash = web3.eth.send_raw_transaction(signed_tx.rawTransaction)
    except Exception as e:
        raise Exception(f"Failed to send the transaction: {e}")
    try:
        tx_receipt = web3.eth.wait_for_transaction_receipt(tx_hash)
    except Exception as e:
        raise Exception(f"Failed to wait for the transaction receipt: {e}")
    return tx_receipt


def decrypt_value(contract_value, user_key):
    # Convert ct to bytes (big-endian)
    byte_array = contract_value.to_bytes(32, byteorder='big')

    # Split ct into two 128-bit arrays r and cipher
    cipher = byte_array[:block_size]
    r = byte_array[block_size:]

    # Decrypt the cipher
    decrypted_message = decrypt(user_key, r, cipher)

    # Print the decrypted cipher
    decrypted_balance = int.from_bytes(decrypted_message, 'big')

    return decrypted_balance




from setuptools import setup

setup(
    name='coti_sdk',
    description='COTI SDK for Privacy Preserving Network',
    version='0.1.7',
    license='Apache2.0',
    author="gmesika-coti@COTI",
    author_email='support@coti.io',
    package_dir={'': '.'},
    url='https://github.com/coti-io/coti-sdk-python',
    keywords='COTI SDK Privacy',
    install_requires=[
        'pycryptodome==3.19.0', 'cryptography==3.4.8', 'eth-keys==0.4.0', 'eth-account==0.10.0', 'web3==6.11.2'
    ],
    python_requires=">=3.9",
)
